@@ -94,6 +94,7 @@ sub dbinfo {
 ##     kbest   => $k,             ##-- return only $k best collocates per date (slice) : default=-1:all
 ##     cutoff  => $cutoff,        ##-- minimum score
 ##     global  => $bool,          ##-- trim profiles globally (vs. locally for each date-slice?) (default=0)
+##     extend  => \%label2gkeys,  ##-- maps slice-labels to selected (packed) group-keys, for extend() method
 ##     ##
 ##     ##-- profiling and debugging parameters
 ##     strings => $bool,          ##-- do/don't stringify item keys (default=do)
@@ -187,6 +188,16 @@ sub profile {
     @{$s2prf->{$_}}{qw(label titles)} = ($_,$groupby->{titles});
   }
 
+  ##-- profile/extend: insert extension keys
+  my $extend = $opts{extend};
+  if ($extend) {
+    my ($slice,$prf,$sxkeys);
+    while (($slice,$prf) = each %$s2prf) {
+      $sxkeys = $extend->{$slice}//{};
+      $prf->{f12}{$_} //= 0 foreach (keys %$sxkeys);
+    }
+  }
+
   ##-- profile: complete slice-wise profiles (pass 2: f2)
   if (!$onepass || !$opts{onepass}) {
     $reldb->vlog($logProfile, "profile(): get frequency profile(s): pass-2");
@@ -221,9 +232,9 @@ sub profile {
 ## Relation API: extend (pass-2 for multi-clients)
 
 ## $mprf = $rel->extend($coldb, %opts)
-##  + get independent f2 frequencies for \%slice2keys = $opts{slice2keys}
-##  + calls $rel->subextend(\%slice2keys, \%opts)
-##  + returns a DiaColloDB::Profile::Multi containing the appropriate f2 entries
+##  + extend f12 and f2 frequencies for \%slice2keys = $opts{slice2keys}
+##  + calls $rel->profile($coldb, %opts,extend=>\%slice2keys_packed)
+##  + returns a DiaColloDB::Profile::Multi containing the appropriate f12 and f2 entries
 sub extend {
   my ($reldb,$coldb,%opts) = @_;
 
@@ -232,40 +243,31 @@ sub extend {
   my $logProfile = $coldb->{logProfile};
 
   ##-- sanity check(s)
-  if (!$opts{slice2keys}) {
-     $reldb->logwarn($coldb->{error}="extend(): no 'slice2keys' parameter specified!");
-     return undef;
-   }
-  elsif (!UNIVERSAL::isa($opts{slice2keys},'HASH')) {
-     $reldb->logwarn($coldb->{error}="extend(): failed to parse 'slice2keys' parameter");
-     return undef;
-   }
+  my $slice2keys = $opts{slice2keys} || $opts{extend};
+  if (!$slice2keys) {
+    $reldb->logwarn($coldb->{error}="extend(): no 'slice2keys' or 'extend' parameter specified!");
+    return undef;
+  }
+  elsif (!UNIVERSAL::isa($slice2keys,'HASH')) {
+    $reldb->logwarn($coldb->{error}="extend(): failed to parse 'slice2keys' or 'extend' parameter");
+    return undef;
+  }
+  delete $opts{slice2keys};
 
-  ##-- subprofile2() requirements
-  my $groupby= $opts{groupby} = $coldb->groupby($opts{groupby});
-  my $a2data = $opts{a2data}  = {map {($_->{a}=>$_)} @{$coldb->attrData($coldb->attrs)}};
-  my $dreq   = $opts{dreq}    = $coldb->parseDateRequest(@opts{qw(date slice fill)});
-
-  ##-- create dummy profiles for completion
-  my $s2prf   = {};
-  my ($slice,$keys);
-  while (($slice,$keys) = each %{$opts{slice2keys}}) {
-    $s2prf->{$slice} = DiaColloDB::Profile->new(label=>$slice, N=>0,f1=>0,f12=>{ map {($_=>0)} @$keys })->stringify($groupby->{s2g});
+  ##-- get packed group-keys (avoid temporary dummy-profiles: they can't handle unknown group-components)
+  my $groupby = $opts{groupby} = $coldb->groupby($opts{groupby});
+  my $s2gx    = $groupby->{s2gx};
+  my ($xslice,$xkeys, $xgkeys,$xkey,$xg, %extend);
+  while (($xslice,$xkeys) = each %$slice2keys) {
+    $xgkeys = $extend{$xslice} = {};
+    foreach $xkey (UNIVERSAL::isa($xkeys,'HASH') ? keys(%$xkeys) : @$xkeys) {
+      next if (!defined($xg = $s2gx->($xkey)));
+      $xgkeys->{$xg} = undef;
+    }
   }
 
-  ##-- get independent f2
-  $reldb->subextend($s2prf, \%opts);
-
-  ##-- collect, stringify & return
-  my $mp = DiaColloDB::Profile::Multi->new(profiles=>[@$s2prf{sort {$a<=>$b} keys %$s2prf}],
-					   titles  =>$groupby->{titles});
-  $reldb->vlog($logProfile, "extend(): stringify");
-  $mp->stringify($groupby->{g2s});
-
-  ##-- trim stringified result profiles (remove unknown elements)
-  foreach my $prf (@{$mp->{profiles}}) {
-    $prf->trim(keep=>$opts{slice2keys}{$prf->{label}});
-  }
+  ##-- guts: dispatch to profile()
+  my $mp = $reldb->profile($coldb, %opts, kbest=>0,kbesta=>0,cutoff=>undef,global=>0,fill=>1, extend=>\%extend);
 
   return $mp;
 }
@@ -396,18 +398,6 @@ sub subprofile2 {
   return $_[1];
 }
 
-##--------------------------------------------------------------
-## Relation API: default: subextend
-
-## \%slice2prf = $rel->subextend(\%slice2prf,\%opts)
-##  + populate f2 frequencies for profiles in \%slice2prf
-##  + %opts: as for subprofile1()
-##  + default implementation just pukes
-sub subextend {
-  my ($rel,$slice2prf,$opts) = @_;
-  $rel->logconfess($opts->{coldb}{error}="subextend() method not supported");
-}
-
 ##--------------------------------------------------------------
 ## Relation API: default: qinfo
 
